"""
Large parts from: https://github.com/samrocketman/gitlab-mirrors/blob/development/lib/manage_gitlab_project.py
"""

import logging
from zope.interface import implementer

from vcs_mirrors.lib.interfaces import IHost

try:
    import gitlab
    from gitlab.exceptions import GitlabCreateError
    GITLAB_AVAILABLE = True
except ImportError:
    GITLAB_AVAILABLE = False
    pass

__all__ = ['__virtual__', 'Host']


def __virtual__():
    if not GITLAB_AVAILABLE:
        logging.warn(
            'Host type "gitlab" isn\'t available, couldn\'t import python-gitlab')
    return GITLAB_AVAILABLE


def _find_matches(objects, kwargs, find_all):
    """Helper function for _add_find_fn. Find objects whose properties
    match all key, value pairs in kwargs.
    Source: https://github.com/doctormo/python-gitlab3/blob/master/gitlab3/__init__.py
    """
    ret = []
    for obj in objects:
        match = True
        # Match all supplied parameters
        for param, val in kwargs.items():
            if not getattr(obj, param) == val:
                match = False
                break
            if match:
                if find_all:
                    ret.append(obj)
                else:
                    return obj
    if not find_all:
        return None

    return ret


@implementer(IHost)
class Host(object):

    TYPE = 'gitlab'

    _settings = None

    def __init__(self, host, settings):
        self._settings = {
            'url': host,
            'ssl_verify': True,
            'public': False,
            'issues_enabled': False,
            'wall_enabled': False,
            'merge_requests_enabled': False,
            'wiki_enabled': False,
            'snippets_enabled': False,
            'use_https': False,
        }

        self._settings.update(settings)

        # pylint: disable=E1101
        self._api = gitlab.Gitlab(
            'https://' + self._settings['url'],
            self._settings['api_key'],
            ssl_verify=self._settings['ssl_verify'],
            api_version=4
        )
        # pylint: enable=E1101
        self._api.auth()

    def _find_group(self, **kwargs):
        groups = self._api.groups.list()
        return _find_matches(groups, kwargs, False)

    def _find_user(self, **kwargs):
        users = self._api.users.list()
        return _find_matches(users, kwargs, False)

    def _find_project(self, **kwargs):
        projects = self._api.projects.list(as_list=True)
        return _find_matches(projects, kwargs, False)

    def create_project(self, source, repo):
        desc = 'Git mirror of %s.' % source
        if self._settings['public']:
            desc = 'Public mirror of %s' % source

        group, name = repo.split('/')

        group_obj = self._find_group(name=group)
        if group_obj is None:
            group_obj = self._find_user(username=group)

        if group_obj is None:
            logging.info('%s: Createing group: "%s"' % (self, group))
            try:
                group_obj = self._api.groups.create({'name': group, 'path': group})
            except GitlabCreateError:
                logging.error('Cannot create group "%s", error: path has already been taken.' % group)
                return False

        project_options = {
            'name': name,
            'description': desc,
            'issues_enabled': str(self._settings['issues_enabled']).lower(),
            'wall_enabled':  str(self._settings['wall_enabled']).lower(),
            'merge_requests_enabled':  str(self._settings['merge_requests_enabled']).lower(),
            'wiki_enabled':  str(self._settings['wiki_enabled']).lower(),
            'snippets_enabled':  str(self._settings['snippets_enabled']).lower(),
            'namespace_id': group_obj.id
        }

        logging.info('%s: Createing project: "%s"' % (self, repo))

        try:
            project = self._api.projects.create(project_options)
        except GitlabCreateError:
            logging.error('Cannot create project "%s", error: path has already been taken.' % repo)
            return False

        if self._settings['use_https']:
            return project.http_url_to_repo

        return project.ssh_url_to_repo

    def get_url(self, repo):
        project = self._find_project(name=repo)
        if project is None:
            return None

        if self._settings['use_https']:
            return project.http_url_to_repo

        return project.ssh_url_to_repo

    def __repr__(self):
        return '<%s(%s)>' % (self.TYPE, self._settings['url'])

    def __str__(self):
        return '%s:%s' % (self.TYPE, self._settings['url'])



import logging

from zope.interface.verify import verifyObject
from zope.interface.exceptions import BrokenImplementation

from vcs_mirrors.lib.interfaces import IHost
from vcs_mirrors.lib.loader import load_hosts
from vcs_mirrors.lib.utils import get_url_host


def configure_argparse(subparser):
    subparser.add_argument('-f', '--force',
                           help='force fetch and push (default False)',
                           dest='force',
                           action='store_true',
                           default=False,
                           required=False)

    subparser.add_argument('-p', '--prune',
                           help='Prune on fetch and push (default False)',
                           dest='prune',
                           action='store_true',
                           default=False,
                           required=False)

    subparser.add_argument('name',
                           help='Internal repository name, for example pcdummy/vcs-mirrors')

    subparser.add_argument('source',
                           help='The source VCS repo URL')

    subparser.add_argument('dest',
                           help='The destination VCS repo URL or <HOST>:<REPO>, if <REPO> is not given we use "name" as REPO')


def execute(config, args):
    if args['name'] in config['repos']:
        logging.fatal('The repository "%s" has already been registered.' % args['name'])
        return 1

    if get_url_host(args['source']) is None:
        logging.fatal('No repository handler for source URL "%s" found.' % args['source'])
        return 1

    dest = args['dest']
    dest_host = None
    dest_split = args['dest'].split(':')
    if dest_split[0] in config['hosts']:
        cfg_host = config['hosts'][dest_split[0]]

        cfg_type = cfg_host['type'].lower()

        hosts = load_hosts()
        for cls in hosts.values():
            if cls.TYPE != cfg_type:
                continue

            dest_host = cls(dest_split[0], cfg_host)
            break

        if dest_host is None:
            logging.fatal('Host type "%s" not implemented.' % cfg_type)
            return 1

        try:
            verifyObject(IHost, dest_host)
        except BrokenImplementation:
            logging.fatal('%r doesn\'t implement IHost correct.', dest_host)
            return 1

        repo = args['name']
        if len(dest_split) > 1:
            repo = dest_split[1]

        url = dest_host.create_project(args['source'], repo)
        if url == False:
            return 1

        dest = url
        logging.info('Destination URL is: "%s".' % url)

    if get_url_host(dest) is None:
        logging.fatal('No repository handler for destination URL "%s" found.' % dest)
        return 1

    repo_cfg = {
        'source': args['source'],
        'dest': dest,
        'force': args['force'],
        'prune': args['prune'],
    }

    config['repos'][args['name']] = repo_cfg

    return 0



import logging
import os
import os.path
import re
import shlex
import subprocess

from zope.interface import implementer

from vcs_mirrors.lib.cd import Cd
from vcs_mirrors.lib.interfaces import IRepo
from vcs_mirrors.lib.utils import run_cmd

__all__ = ['Repo']


@implementer(IRepo)
class Repo(object):
    TYPE = 'git'

    @staticmethod
    def get_host(url):
        """
        Returns the hostname when the url is a git host else None

        Tested urls:
            - https://github.com/pcdummy/vcs-mirrors.git
            - https://git.lxch.eu/vcs-mirrors.git
            - ssh://user@server/project.git
            - user@server:project.git
            - git://git.proxmox.com/git/aab.git
        """
        url = str(url)

        # https://github.com/pcdummy/vcs-mirrors.git
        # https://git.lxch.eu/vcs-mirrors.git
        match = re.search(r'https://([-\d\w_\.]+)/.*\.git$', url)
        if match:
            return match.group(1)

        # ssh://user@server/project.git
        match = re.search(
            r'(ssh://)?[-\d\w_\.]+@{1}([-\d\w_\.]+)[:/]{1}.*\.git$', url)
        if match:
            return match.group(2)

        # git://git.lxch.eu/git/aab.git
        match = re.search(r'git://([-\d\w_\.]+)/.*\.git$', url)
        if match:
            return match.group(1)

        return None

    _config = None

    def _after_clone(self):
        pass

    def __init__(self, config):
        self._config = config

    def fetch(self, repo):
        logging.debug('%s: Fetching repository "%s"' % (self, repo))

        repo_config = self._config['repos'][repo]
        force = False
        if 'force' in repo_config:
            force = repo_config['force']

        prune = False
        if 'prune' in repo_config:
            prune = repo_config['prune']

        repo_dir = os.path.join(self._config['settings']['local_path'], repo)
        repo_dir_exists = True
        if not os.path.exists(repo_dir):
            repo_dir_exists = False
            logging.debug('%s: Makedirs: "%s"' % (self, repo_dir))
            os.makedirs(repo_dir, mode=0o700)

        logging.debug('%s: cd %s' % (self, os.path.abspath(repo_dir)))
        with Cd(repo_dir):
            if not repo_dir_exists:
                run_cmd(['git', 'clone', '--mirror', shlex.quote(repo_config['source']), '.'])
                #run_cmd(['git', 'remote', 'rename', 'origin', 'source'])
                self._after_clone()
            else:
                has_repo = run_cmd(['git', 'remote', 'get-url', 'origin'], False).returncode == 0
                if not has_repo:
                    run_cmd(['git', 'remote', 'add', 'origin', shlex.quote(repo_config['source'])])

                args = ['git', 'fetch']
                if force:
                    args.append('--force')
                if prune:
                    args.append('--prune')
                args.append('origin')

                run_cmd(args)

        return True

    def push(self, repo):
        logging.debug('%s: Pushing to repository "%s"' % (self, repo))

        repo_dir = os.path.join(self._config['settings']['local_path'], repo)
        if not os.path.exists(repo_dir):
            logging.error('%s: Can\'t push repo "%s", it does not exists localy.' % (self, repo))

        repo_config = self._config['repos'][repo]
        force = False
        if 'force' in repo_config:
            force = repo_config['force']

        prune = False
        if 'prune' in repo_config:
            prune = repo_config['prune']

        with Cd(repo_dir):
            has_repo = run_cmd(['git', 'remote', 'get-url', 'dest'], False).returncode == 0
            if not has_repo:
                run_cmd(['git', 'remote', 'add', 'dest', shlex.quote(repo_config['dest'])])
                run_cmd(['git', 'config', '--add', 'remote.dest.push', "+refs/heads/*:refs/heads/*"])
                run_cmd(['git', 'config', '--add', 'remote.dest.push', "+refs/tags/*:refs/tags/*"])
                run_cmd(['git', 'config', 'remote.dest.mirror', 'true'])

                if repo_config['dest'][0:4] == 'http':
                    run_cmd(['git', 'config', 'credential.helper', 'store'])

            args = ['git', 'push']
            if force:
                args.append('--force')
            if prune:
                args.append('--prune')
            args.extend(['dest'])

            run_cmd(args)

        return True

    def __str__(self):
        return self.TYPE

    def __repr__(self):
        return '<repo(%s)>' % self.TYPE



import logging

from vcs_mirrors.lib.utils import sync_one


def configure_argparse(subparser):
    pass

def execute(config, args):
    result = 0
    for repo in config['repos'].keys():
        logging.info('Syncing "%s"' % repo)
        if sync_one(repo, config) == False:
            result = 1

    return result



import logging

from zope.interface import implementer

from vcs_mirrors.lib.interfaces import IHost

try:
    from github import Github
    from github.GithubException import GithubException
    from github.GithubException import UnknownObjectException
    GITHUB_AVAILABLE = True
except ImportError:
    GITHUB_AVAILABLE = False
    pass

__all__ = ['__virtual__', 'Host']


def __virtual__():
    if not GITHUB_AVAILABLE:
        logging.warn(
            'Host type "github" isn\'t available, couldn\'t import pygithub')
    return GITHUB_AVAILABLE


@implementer(IHost)
class Host(object):

    TYPE = 'github'

    _settings = None

    def __init__(self, host, settings):
        self._settings = {
            'public': True,
            'issues_enabled': True,
            'wiki_enabled': True,
            'downloads_enabled': True,
            'projects_enabled': True,
            'use_https': False,
        }

        self._settings.update(settings)

        self._api = Github(self._settings['api_key'])

    def create_project(self, source, repo):
        desc = 'Git mirror of %s.' % source
        if self._settings['public']:
            desc = 'Public mirror of %s' % source

        org, name = repo.split('/')
        g_user = self._api.get_user()

        logging.info('%s: Createing project: "%s"' % (self, repo))
        g_repo = None
        if org == g_user.login:
            try:
                g_repo = g_user.create_repo(
                    name,
                    description=desc,
                    private=not self._settings['public'],
                    has_issues=self._settings['issues_enabled'],
                    has_wiki=self._settings['wiki_enabled'],
                    has_downloads=self._settings['downloads_enabled'],
                    has_projects=self._settings['projects_enabled'],
                )
            except GithubException:
                return False

        else:
            # Find org
            try:
                g_org = self._api.get_organization(org)
            except UnknownObjectException:
                return False

            # Create repo
            try:
                g_repo = g_org.create_repo(
                    name,
                    description=desc,
                    private=not self._settings['public'],
                    has_issues=self._settings['issues_enabled'],
                    has_wiki=self._settings['wiki_enabled'],
                    has_downloads=self._settings['downloads_enabled'],
                    has_projects=self._settings['projects_enabled'],
                )
            except GithubException:
                return False

        if self._settings['use_https']:
            return g_repo.clone_url

        return g_repo.ssh_url

    def get_url(self, repo):
        g_repo = self._api.get_repo(repo)

        if self._settings['use_https']:
            return g_repo.clone_url

        return g_repo.ssh_url

    def __repr__(self):
        return '<%s>' % self.TYPE

    def __str__(self):
        return '%s' % self.TYPE



import re

from vcs_mirrors.lib.utils import run_cmd
from vcs_mirrors.repo.git import Repo as GitRepo


class Repo(GitRepo):
    TYPE = 'bazaar'

    @staticmethod
    def get_host(url):
        """
        Returns the hostname when the url is a git host else None

        Tested urls:
            - bzr::bzr://bzr.savannah.gnu.org/emacs/trunk
            - bzr::bzr://bzr.savannah.gnu.org/emacs
            - bzr::lp:ubuntu/hello
            - bzr::lp:bzr
            - bzr::sftp://bill@mary-laptop/cool-repo/cool-trunk
        """
        url = str(url).lower()
        match = re.search(r'bzr::([-\d\w_\.]+):(//)?([\-0-9a-z_\.]+@+)?([\-0-9a-z_\.]+)(/.+)?', url)
        if match:
            if match.group(1) == 'lp':
                return 'lp'
            else:
                return match.group(2)

        return None

    def _after_clone(self):
        run_cmd(['git', 'gc', '--aggressive'])

        super()._after_clone()



import logging

from vcs_mirrors.lib.utils import sync_one


def configure_argparse(subparser):
    subparser.add_argument('name',
                           help='Internal repository name')


def execute(config, args):
    logging.info('Syncing "%s"' % args['name'])
    if sync_one(args['name'], config) == False:
        return 1

    return 0



from zope.interface import Attribute, Interface


# pylint: disable=E0239,E0213
class IHost(Interface):
    """
    Host interface, all Hosts MUST implement this
    """

    TYPE = Attribute("""Type of the Host (normaly lowercase of module name)""")

    def create_project(source, repo):
        """
        Create the repo "repo" on the host
        Returns the URL for the created project or False on error.

        :return: url or False
        """

    def get_url(repo):
        """
        Returns the URL for the given repo
        and None if not found.

        :return: url or None
        """


class IRepo(Interface):
    """
    Repo interface, all Repos MUST implement this
    """

    TYPE = Attribute("""Type of the Repo (normaly lowercase of module name)""")

    def fetch(repo):
        """
        Mirrors the source url to the internal store.

        :param repo: Internal repo name
        :return: boolean
        """

    def push(repo):
        """
        Push from the internal store to the destination.

        :param repo: Internal repo name
        :return: boolean
        """
# pylint: enable=E0239,E0213



import logging
import subprocess

from vcs_mirrors.lib.loader import load_repos


def get_url_host(url):
    repos = load_repos()

    for repo in repos.values():
        host = repo.get_host(url)
        if host is not None:
            return host

    return None


def get_repo_for_url(url, config):
    repos = load_repos()

    result = None
    for repo in repos.values():
        host = repo.get_host(url)
        if host is not None:
            result = repo(config)
            break

    return result


def sync_one(name, config):
    if name not in config['repos']:
        logging.fatal('Unknown repository "%s" given.' % name)
        return False

    repo_config = config['repos'][name]

    source_repo = get_repo_for_url(repo_config['source'], config)
    if source_repo is None:
        logging.fatal(
            'No repository handler for source URL "%s" found.' % repo_config['source'])
        return False

    dest_repo = get_repo_for_url(repo_config['dest'], config)
    if dest_repo is None:
        logging.fatal(
            'No repository handler for source URL "%s" found.' % repo_config['dest'])
        return False

    if source_repo.fetch(name) == False:
        return False

    if dest_repo.push(name) == False:
        return False

    return True

def run_cmd(args, log_error=True):
    logging.debug('Run: %s' % " ".join(args))
    result = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if log_error:
        if result.returncode != 0:
            logging.error(result.stderr.decode('unicode_escape'))
    return result



import importlib
import pkgutil

import logging
import vcs_mirrors.command
import vcs_mirrors.host
import vcs_mirrors.repo
from vcs_mirrors.lib.interfaces import IHost, IRepo

REPOS = None
HOSTS = None
COMMANDS = None


def _load_all(package):
    """
    https://stackoverflow.com/a/1707786/3368468
    """
    result = {}
    for _, modname, ispkg in pkgutil.iter_modules(package.__path__):
        if ispkg == False:
            module = importlib.import_module(
                package.__name__ + '.' + modname, package)
            result[modname] = module
    return result


def load_commands():
    global COMMANDS
    if COMMANDS is not None:
        return COMMANDS

    COMMANDS = _load_all(vcs_mirrors.command)
    return COMMANDS


def load_repos():
    global REPOS
    if REPOS is not None:
        return REPOS

    REPOS = {}
    for k, m in _load_all(vcs_mirrors.repo).items():
        # pylint: disable=E1120
        if IRepo.implementedBy(m.Repo):
            REPOS[k] = m.Repo
        else:
            logging.error('Repo "%s" doesn\'t implement IRepo' % k)
        # pylint: enable=E1120

    return REPOS


def load_hosts():
    global HOSTS
    if HOSTS is not None:
        return HOSTS

    HOSTS = {}
    for k, m in _load_all(vcs_mirrors.host).items():
        if not m.__virtual__():
            continue

        # pylint: disable=E1120
        if IHost.implementedBy(m.Host):
            HOSTS[k] = m.Host
        else:
            logging.error('Host "%s" doesn\'t implement IHost' % k)
        # pylint: enable=E1120

    return HOSTS



import ruamel.yaml


def load_config(path):
    with open(path, 'r') as fp:
        return ruamel.yaml.round_trip_load(fp.read())


def save_config(config, path):
    with open(path, 'w') as fp:
        ruamel.yaml.round_trip_dump(config, fp)



import codecs
import os
from setuptools import setup, find_packages

# abspath here because setup.py may be __main__, in which case
# __file__ is not guaranteed to be absolute
here = os.path.abspath(os.path.dirname(__file__))


def read_file(filename):
    """Open a related file and return its content."""
    with codecs.open(os.path.join(here, filename), encoding='utf-8') as f:
        content = f.read()
    return content


README = read_file('README.rst')
CHANGELOG = read_file('CHANGELOG.rst')
CONTRIBUTORS = read_file('CONTRIBUTORS.rst')

REQUIREMENTS = [
    'ruamel.yaml',
    'zope.interface',
]

GITLAB_REQUIRES = [
    'python-gitlab',
]

GITHUB_REQUIRES = [
    'pygithub',
]

DEVELOPMENT_REQUIRES = [
    'pylint',
    'autopep8',
    'flake8',
    'ipython',
    'zest.releaser',
]

DEPENDENCY_LINKS = []

ENTRY_POINTS = {
    'console_scripts': [
        'vcs-mirrors = vcs_mirrors.scripts.main:main'
    ]
}

setup(name='vcs_mirrors',
      version='0.0.7',
      description='',
      long_description='{}\n\n{}\n\n{}'.format(README, CHANGELOG, CONTRIBUTORS),
      license='MIT',
      classifiers=[
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'License :: OSI Approved :: MIT License',
          'Topic :: Software Development :: Version Control',
          'Topic :: Software Development :: Version Control :: Bazaar',
          'Topic :: Software Development :: Version Control :: Git',
      ],
      keywords='Console VCS mirror vcs-mirrors',
      author='René Jochum',
      author_email='rene@jochums.at',
      url='https://git.lxch.eu/pcdummy/vcs-mirrors',
      packages=find_packages(),
      package_data={'': ['*.rst', '*.py', '*.yaml']},
      include_package_data=True,
      zip_safe=False,
      install_requires=REQUIREMENTS,
      extras_require={
          'gitlab': GITLAB_REQUIRES,
          'github': GITHUB_REQUIRES,
          'development': DEVELOPMENT_REQUIRES,
      },
      dependency_links=DEPENDENCY_LINKS,
      entry_points=ENTRY_POINTS)



import argparse
import os
import sys
import logging

from vcs_mirrors.lib.config import load_config
from vcs_mirrors.lib.config import save_config
from vcs_mirrors.lib.loader import load_commands


DEFAULT_CONFIG_FILE = os.getenv('VCS_MIRROR_CONFIG', 'vcs-mirrors.yaml')
DEFAULT_LOG_LEVEL = logging.INFO
DEFAULT_LOG_FORMAT = '%(levelname)-8.8s  %(message)s'


def main(args=None):
    """The main routine."""
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser(
        description='vcs-mirror Command-Line Interface')

    subparsers = parser.add_subparsers(title='subcommands',
                                       description='Main vcs-mirror CLI commands',
                                       dest='subcommand',
                                       help='Choose and run with --help')
    subparsers.required = True

    cmds = load_commands()

    for name, cmd in cmds.items():
        subparser = subparsers.add_parser(name)
        subparser.set_defaults(which=name)

        subparser.add_argument('--config',
                               help='Application configuration file',
                               dest='yaml_file',
                               required=False,
                               default=DEFAULT_CONFIG_FILE)

        subparser.add_argument('-q', '--quiet', action='store_const',
                               const=logging.CRITICAL, dest='verbosity',
                               help='Show only critical errors.')

        subparser.add_argument('-v', '--debug', action='store_const',
                               const=logging.DEBUG, dest='verbosity',
                               help='Show all messages, including debug messages.')

        cmd.configure_argparse(subparser)


    # Parse command-line arguments
    parsed_args = vars(parser.parse_args(args))

    # Initialize logging from
    level = parsed_args.get('verbosity') or DEFAULT_LOG_LEVEL
    logging.basicConfig(level=level, format=DEFAULT_LOG_FORMAT)

    config = load_config(parsed_args['yaml_file'])
    if config['repos'] is None:
        config['repos'] = {}

    # Execute the command
    which_command = parsed_args['which']
    result = cmds[which_command].execute(config, parsed_args)
    if result != 0:
        sys.exit(result)

    save_config(config, parsed_args['yaml_file'])
